@@ -1,6 +1,6 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /cvsroot/pgpool/pgpool/main.c,v 1.12 2006/05/12 03:46:15 t-ishii Exp $
+ * $Header: /cvsroot/pgpool/pgpool/main.c,v 1.13 2006/08/02 08:27:29 t-ishii Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
@@ -178,8 +178,12 @@ int main(int argc, char **argv)
 	pool_config.current_backend_port = pool_config.backend_port;
 
 	/* set load balance weight */
-	weight_master = (RAND_MAX) * (pool_config.weight_master /
-						   (pool_config.weight_master + pool_config.weight_secondary));
+	if (pool_config.weight_master <= 0.0)
+		weight_master = 0;
+	else
+		weight_master = (RAND_MAX) * (pool_config.weight_master /
+									  (pool_config.weight_master + pool_config.weight_secondary));
+
 	pool_debug("weight: %ld", weight_master);
 
 	/*
