@@ -1,7 +1,7 @@
 /* -*-pgsql-c-*- */
 /*
  *
- * $Header: /cvsroot/pgpool/pgpool/pool.h,v 1.10 2006/06/04 10:03:30 t-ishii Exp $
+ * $Header: /cvsroot/pgpool/pgpool/pool.h,v 1.11 2006/08/30 01:11:47 t-ishii Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
@@ -290,6 +290,8 @@ extern void pool_discard_cp(char *user, char *database, int protoMajor);
 
 extern POOL_STATUS ErrorResponse(POOL_CONNECTION *frontend, 
 								  POOL_CONNECTION_POOL *backend);
+extern POOL_STATUS ErrorResponse2(POOL_CONNECTION *frontend, 
+								  POOL_CONNECTION_POOL *backend);
 
 extern POOL_STATUS NoticeResponse(POOL_CONNECTION *frontend, 
 								  POOL_CONNECTION_POOL *backend);
@@ -311,6 +313,7 @@ extern void pool_send_frontend_exits(POOL_CONNECTION_POOL *backend);
 extern int pool_read_message_length(POOL_CONNECTION_POOL *cp);
 extern int *pool_read_message_length2(POOL_CONNECTION_POOL *cp);
 extern signed char pool_read_kind(POOL_CONNECTION_POOL *cp);
+extern signed char pool_read_kind2(POOL_CONNECTION_POOL *cp);
 
 extern POOL_STATUS SimpleForwardToFrontend(char kind, POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend);
 extern POOL_STATUS SimpleForwardToBackend(char kind, POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend);
