@@ -1,11 +1,11 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /cvsroot/pgpool/pgpool/pool_auth.c,v 1.7 2005/11/11 11:51:04 t-ishii Exp $
+ * $Header: /cvsroot/pgpool/pgpool/pool_auth.c,v 1.10 2006/08/30 01:11:48 t-ishii Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2005	Tatsuo Ishii
+ * Copyright (c) 2003-2006	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -55,7 +55,15 @@ int pool_do_auth(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *cp)
 
 	protoMajor = MAJOR(cp);
 
-	kind = pool_read_kind(cp);
+        if (protoMajor == PROTO_MAJOR_V3)
+        {
+                kind = pool_read_kind2(cp);
+        }
+        else
+        {
+                kind = pool_read_kind(cp);
+        }
+
 	if (kind < 0)
 	{
 		return -1;
@@ -70,7 +78,14 @@ int pool_do_auth(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *cp)
 		 * will negotiate again using v2 protocol.
 		 */
 		pool_log("pool_do_auth: maybe protocol version mismatch (current version %d)", protoMajor);
-		ErrorResponse(frontend, cp);
+		if (protoMajor == PROTO_MAJOR_V3)
+		{
+			ErrorResponse2(frontend, cp);
+		}
+		else
+		{
+			ErrorResponse(frontend, cp);
+		}
 		return -1;
 	}
 	else if (kind != 'R')
@@ -407,6 +422,9 @@ static POOL_STATUS pool_send_auth_ok(POOL_CONNECTION *frontend, int pid, int key
 		len = htonl(12);
 		pool_write(frontend, &len, sizeof(len));
 	}
+
+	pool_debug("pool_send_auth_ok: send pid %d to frontend", ntohl(pid));
+
 	pool_write(frontend, &pid, sizeof(pid));
 	if (pool_write_and_flush(frontend, &key, sizeof(key)) < 0)
 	{
@@ -1032,3 +1050,38 @@ signed char pool_read_kind(POOL_CONNECTION_POOL *cp)
 
 	return kind;
 }
+
+signed char pool_read_kind2(POOL_CONNECTION_POOL *cp)
+{
+	int status;
+	char kind, kind1;
+	char *buf;
+
+	buf = pool_read2(MASTER(cp), sizeof(kind));
+	if (buf == NULL)
+	{
+		pool_error("read_kind2: error while reading message kind");
+		return -1;
+	}
+
+	kind = *buf;
+
+	if (DUAL_MODE)
+	{
+		status = pool_read(SECONDARY(cp), &kind1, sizeof(kind1));
+		if (status < 0)
+		{
+			pool_error("read_kind2: error while reading message kind from secondary backend");
+			return -1;
+		}
+
+		if (kind != kind1)
+		{
+			pool_error("read_kind2: kind does not match between backends master(%d) secondary(%d)",
+					   kind, kind1);
+			return -1;
+		}
+	}
+
+	return kind;
+}
